/*
    KSysGuard, the KDE System Guard
           
        Copyright (c) 2004 Stefan Teleman <stefan.teleman@gmail.com>
        *** several functions used in here have been borrowed ***
        *** from the equivalent FreeBSD translation unit. ***
    
    This program is free software; you can redistribute it and/or
    modify it under the terms of version 2 of the GNU General Public
    License as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/mnttab.h>
#include <sys/mntent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/statvfs.h>
#include <sys/param.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>

#include "Command.h"
#include "ccont.h"
#include "DiskInfo.h"
#include "ksysguardd.h"

typedef struct StatVFSInfo
{
  unsigned long blockSize;
  unsigned long frSize;
#if defined(_LP64) || _FILE_OFFSET_BITS == 32
  ulong_t totalBlocks;
  ulong_t freeBlocks;
  ulong_t usedBlocks;
#elif _FILE_OFFSET_BITS == 64
  u_longlong_t totalBlocks;
  u_longlong_t freeBlocks;
  u_longlong_t usedBlocks;
#endif
  double usedPercent;
} StatVFSInfo;

typedef struct DiskInfo
{
  StatVFSInfo vfsInfo;
  char device[MAXPATHLEN + sizeof(char)];
  char mountPoint[MAXPATHLEN + sizeof(char)];
  char fsType[FSTYPSZ + sizeof(char)];
} DiskInfo;

static DiskInfo* diskInfoArray = 0L;
static struct SensorModul* DiskInfoSM = 0L;
static unsigned int numberOfMountedDevices = 0;
static char device[MAXPATHLEN + sizeof(char)];

extern int errno;

static char*
getMountPoint (const char* cmd)
{
  char* ptr;
  char format[32];

  (void) memset (device, '\0', sizeof (device));
  (void) memset (format, '\0', 32);
  (void) sprintf (format, "partitions%%%us", (uint32_t) MAXPATHLEN);
  (void) sscanf (cmd, format, device);

  ptr = (char *) rindex (device, '/');
  if (ptr)
  {
    *ptr = '\0';
    return device;
  }
  return 0L;
}

static void
copyMountInfo (const struct extmnttab* mp, DiskInfo* dp)
{
  if ((0L == mp) || (0L == dp))
    return;

  (void) strcpy (dp->device, mp->mnt_special);
  (void) strcpy (dp->mountPoint, mp->mnt_mountp);
  (void) strcpy (dp->fsType, mp->mnt_fstype);
}

static void
copyVFSInfo (const struct statvfs* vfsp, StatVFSInfo* sp)
{
  if ((0L == vfsp) || (0L == sp))
    return;

  sp->blockSize = vfsp->f_bsize;
  sp->frSize = vfsp->f_frsize;
  sp->totalBlocks = (unsigned long) vfsp->f_blocks;
  sp->freeBlocks = (unsigned long) vfsp->f_bavail;
  sp->usedBlocks = (unsigned long) sp->totalBlocks - sp->freeBlocks;
  sp->usedPercent = ((double) sp->totalBlocks - (double) sp->freeBlocks) / (double) sp->totalBlocks;
  sp->usedPercent = sp->usedPercent * 100;
}

static void
printMntTab (struct extmnttab* mp)
{
  (void) fprintf (stderr, "mnt_special: %s\n", mp->mnt_special);
  (void) fprintf (stderr, "mnt_mountp: %s\n", mp->mnt_mountp);
  (void) fprintf (stderr, "mnt_fstype: %s\n", mp->mnt_fstype);
  (void) fprintf (stderr, "mnt_mntopts: %s\n", mp->mnt_mntopts);
  (void) fprintf (stderr, "mnt_time: %s\n", mp->mnt_time);
  (void) fprintf (stderr, "mnt_major: %u\n", mp->mnt_major);
  (void) fprintf (stderr, "mnt_minor: %u\n", mp->mnt_minor);
}

static void
printVfsInfo (struct statvfs* vfsp)
{
  (void) fprintf (stderr, "f_bsize = %u\n", vfsp->f_bsize);
  (void) fprintf (stderr, "f_frsize = %u\n", vfsp->f_frsize);
  (void) fprintf (stderr, "f_blocks = %llu\n", vfsp->f_blocks);
  (void) fprintf (stderr, "f_bavail = %llu\n", vfsp->f_bavail);
  (void) fprintf (stderr, "f_bfree = %llu\n", vfsp->f_blocks - vfsp->f_bavail);
}

static unsigned int
getMountPointInfo (void)
{
  struct extmnttab mp;
  struct statvfs vp;
  FILE* fp = 0L;
  uint32_t i;
  int ret = 0;
  char line[1024];

  if ((fp = fopen (MNTTAB, "r")) == (FILE *) 0L)
  {
    perror("fopen"), strerror(errno);
    return 0L;
  }
  numberOfMountedDevices = 0;
  free (diskInfoArray);
  diskInfoArray = 0L;
  (void) memset ((void *) &mp, '\0', sizeof (struct extmnttab));

  while ((ret = getextmntent (fp, &mp, MNT_LINE_MAX)) == 0)
  {
    if ((0 == strcmp (mp.mnt_fstype, "ufs")) || (0 == strcmp (mp.mnt_fstype, "nfs")))
    {
      if (0 == strcmp (mp.mnt_mountp, "/vol"))
        continue;

      ++numberOfMountedDevices;
      diskInfoArray = (DiskInfo *) realloc (diskInfoArray, numberOfMountedDevices * sizeof (struct DiskInfo));
      copyMountInfo (&mp, &(diskInfoArray[numberOfMountedDevices -1]));
    }
  }
  resetmnttab (fp);
  fclose (fp);

  (void) memset (&vp, '\0', sizeof (struct statvfs));

  for (i = 0; i < numberOfMountedDevices; i++)
  {
    if (0 == statvfs (diskInfoArray[i].mountPoint, &vp))
      copyVFSInfo (&vp, &(diskInfoArray[i].vfsInfo));
  }
  return numberOfMountedDevices;
}

void
initDiskInfo (struct SensorModul* sm)
{
  char monitor[MAXPATHLEN + 12 * sizeof(char)];
  uint32_t i;

  DiskInfoSM = sm;
  updateDiskInfo();

  registerMonitor("partitions/list", "listview", printDiskInfo, printDiskInfoInfo, DiskInfoSM);

  for (i = 0; i < numberOfMountedDevices; i++)
  {
    (void) snprintf (monitor, sizeof(monitor), "partitions%s/usedspace", ((strcmp (diskInfoArray[i].mountPoint, "/") == 0) ? "/root" : diskInfoArray[i].mountPoint));
    registerMonitor (monitor, "integer", printDiskInfoUsed, printDiskInfoUsedInfo, DiskInfoSM);
    (void) snprintf (monitor, sizeof(monitor), "partitions%s/freespace", ((strcmp (diskInfoArray[i].mountPoint, "/") == 0) ? "/root" : diskInfoArray[i].mountPoint));
    registerMonitor (monitor, "integer", printDiskInfoFree, printDiskInfoFreeInfo, DiskInfoSM);
    (void) snprintf (monitor, sizeof(monitor), "partitions%s/filllevel", ((strcmp (diskInfoArray[i].mountPoint, "/") == 0) ? "/root" : diskInfoArray[i].mountPoint));
    registerMonitor (monitor, "float", printDiskInfoPercent, printDiskInfoPercentInfo, DiskInfoSM);
  }
}

void
exitDiskInfo (void)
{
  char monitor[MAXPATHLEN + 12 * sizeof(char)];
  uint32_t i;

  removeMonitor("partitions/list");

  for (i = 0; i < numberOfMountedDevices; i++)
  {
    (void) snprintf (monitor, sizeof(monitor), "partitions%s/usedspace", ((strcmp (diskInfoArray[i].mountPoint, "/") == 0) ? "/root" : diskInfoArray[i].mountPoint));
    removeMonitor (monitor);
    (void) snprintf (monitor, sizeof(monitor), "partitions%s/freespace", ((strcmp (diskInfoArray[i].mountPoint, "/") == 0) ? "/root" : diskInfoArray[i].mountPoint));
    removeMonitor (monitor);
    (void) snprintf (monitor, sizeof(monitor), "partitions%s/filllevel", ((strcmp (diskInfoArray[i].mountPoint, "/") == 0) ? "/root" : diskInfoArray[i].mountPoint));
    removeMonitor (monitor);
    free (diskInfoArray);
    diskInfoArray = 0L;
  }
}

void
checkDiskInfo (void)
{
  struct extmnttab mp;
  unsigned int ctr = 0;
  int ret = 0;
  FILE* fp = 0L;
  uint32_t i;
  char line[1024];

  if ((fp = fopen (MNTTAB, "r")) == (FILE *) 0L)
  {
    perror("fopen"), strerror(errno);
    return;
  }
  (void) memset ((void *) &mp, '\0', sizeof (struct extmnttab));

  while ((ret = getextmntent (fp, &mp, MNT_LINE_MAX)) == 0)
  {
    if ((0 == strcmp (mp.mnt_fstype, "ufs")) || (0 == strcmp (mp.mnt_fstype, "nfs")))
    {
      if (0 == strcmp (mp.mnt_mountp, "/vol"))
        continue;

      ++ctr;
    }
  }
  resetmnttab (fp);
  fclose (fp);

  if (ctr != numberOfMountedDevices)
  {
    exitDiskInfo();
    initDiskInfo (DiskInfoSM);
  }
}

int
updateDiskInfo (void)
{
  if (getMountPointInfo() > 0)
    return 0;

  return -1;
}

void
printDiskInfo (const char* cmd)
{
  uint32_t i;

  for (i = 0; i < numberOfMountedDevices; i++)
  {
    (void) fprintf (CurrentClient, "%s\t%llu\t%llu\t%llu\t%.2lf\t%s\n",
                    diskInfoArray[i].device,
                    diskInfoArray[i].vfsInfo.totalBlocks,
                    diskInfoArray[i].vfsInfo.usedBlocks,
                    diskInfoArray[i].vfsInfo.freeBlocks,
                    diskInfoArray[i].vfsInfo.usedPercent,
                    diskInfoArray[i].mountPoint);
  }
  (void) fprintf (CurrentClient, "\n");
}

void
printDiskInfoInfo (const char* cmd)
{
  (void) fprintf(CurrentClient, "Device\tBlocks\tUsed\tAvailable\tUsed %%\tMountPoint\nM\tD\tD\tD\tf\ts\n");
}

void
printDiskInfoUsed (const char* cmd)
{
  uint32_t i;

  char* mountPoint = getMountPoint (cmd);

  if (mountPoint)
  {
    for (i = 0; i < numberOfMountedDevices; i++)
    {
      if (0 == strcmp (mountPoint, diskInfoArray[i].mountPoint))
        (void) fprintf (CurrentClient, "%llu\n", diskInfoArray[i].vfsInfo.usedBlocks);
    }
  }
  (void) fprintf (CurrentClient, "\n");
}

void
printDiskInfoUsedInfo (const char* cmd)
{
  (void) fprintf (CurrentClient, "Used Blocks\t0\t-\tBlocks\n");
}

void
printDiskInfoFree (const char* cmd)
{
  uint32_t i;

  char* mountPoint = getMountPoint (cmd);
  if (mountPoint)
  {
    for (i = 0; i < numberOfMountedDevices; i++)
    {
      if (0 == strcmp (mountPoint, diskInfoArray[i].mountPoint))
        (void) fprintf (CurrentClient, "%llu\n", diskInfoArray[i].vfsInfo.freeBlocks);
    }
  }
  (void) fprintf (CurrentClient, "\n");
}

void
printDiskInfoFreeInfo (const char* cmd)
{
  (void) fprintf (CurrentClient, "Free Blocks\t0\t-\tBlocks\n");
}

void
printDiskInfoPercent (const char* cmd)
{
  uint32_t i;

  char* mountPoint = getMountPoint (cmd);
  if (mountPoint)
  {
    for (i = 0; i < numberOfMountedDevices; i++)
    {
      if (0 == strcmp (mountPoint, diskInfoArray[i].mountPoint))
        (void) fprintf (CurrentClient, "%.2lf\n", diskInfoArray[i].vfsInfo.usedPercent);
    }
  }
}

void
printDiskInfoPercentInfo (const char* cmd)
{
  (void) fprintf (CurrentClient, "Used Blocks\t0\t100\t%%\n");
}

// vim: ts=2 sw=2 et


