/*
  KSysGuard, the KDE System Guard
  
  Copyright (c) 2005 Stefan Teleman <steleman@acm.org>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of version 2 of the GNU General Public
  License as published by the Free Software Foundation.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
*/

#include <config.h>

#include <stdio.h>
#include <sys/types.h>
#include <sys/dkio.h>
#include <sys/stat.h>
#include <sys/mntent.h>
#include <sys/mnttab.h>
#include <sys/statvfs.h>
#include <sys/utsname.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <inttypes.h>
#include <kstat.h>
#include <fcntl.h>
#include <sys/fcntl.h>
#include <sys/param.h>
#include <sys/mkdev.h>
#include <sys/sysinfo.h>
#include <devid.h>
#include <libdevinfo.h>
#include <dirent.h>

#include <errno.h>

#include "Command.h"
#include "ccont.h"
#include "SCSIStat.h"
#include "ksysguardd.h"

// DDI/DKI stuff
typedef struct DDIInfo
{
  int instance;
  char driver[8];
  char physicalPath[MAXPATHLEN];
} DDIInfo;

// KStat stuff
typedef struct DiskIOInfo
{
  kstat_io_t old;
  kstat_io_t new;
} DiskIOInfo;

// SCSI stuff
typedef struct DiskSCSIInfo
{
    int instance;
    int kid;
    int lun;
    int target;
    dev_t dev;
    major_t major;
    minor_t minor;
    char kstatName[KSTAT_STRLEN];
    char kstatClass[KSTAT_STRLEN];
    char vendorName[32];
    char productId[64];
    char revision[32];
    char ddiDeviceName[MAXPATHLEN];
    char specType[16];
    char nodeType[64];
    char svr4Device[MAXPATHLEN];
    char scsiSlice[32];
    char diMinorName[2];
    kstat_io_t kstatIO;
} DiskSCSIInfo;

typedef struct MnttabInfo
{
    char device[MAXPATHLEN + sizeof(char)];
    char mountPoint[MAXPATHLEN + sizeof(char)];
    char fsType[FSTYPSZ + sizeof(char)];
    int32_t major;
    int32_t minor;
} MnttabInfo;


// NFS stuff
typedef struct NFSMountInfo
{
    char mnt_special[MAXPATHLEN];
    char mnt_mountp[MAXPATHLEN];
    char mnt_mntopts[MNT_LINE_MAX];
    uint_t mnt_major;
    uint_t mnt_minor;
} NFSMountInfo;

typedef struct UTSName
{
  int ok;
  struct utsname uName;
} UTSName;

static char driver[8];
static UTSName utsName;
// static DiskInfo* diskInfoArray = 0L;
static struct SensorModul* DiskStatSM = 0L;
// static unsigned int numSCSIInfo = 0;
static unsigned int numberOfMnttabEntries = 0;
static char device[MAXPATHLEN + sizeof(char)];

static DiskSCSIInfo* scsiInfoArray = 0L;
static DDIInfo* ddiInfoPtr = 0L;
static MnttabInfo* mnttabInfoPtr = 0L;
static int* instancePtr = 0L;
static size_t numDDIInfo = 0;
static size_t numSCSIInfo = 0;
static size_t numInstances = 0;
static int hasDriver = 0;
static struct timeval tv;

static void fillDefaultValues (DiskSCSIInfo* dp);
static void setUTSName (void);
static void setDriverName (void);
static void copyDiskSCSIInfo (const kstat_t* ksp, const kstat_io_t* kip, DiskSCSIInfo* dp);
static void resetMemory (void);
static int getMnttabInfo (void);
static int getSCSIStatInfo (void);
static char* getSCSIDevice (const char* cmd);
static int updateKStatList (void);
static int updateDeviceList (void);
static int dumpDeviceNode (di_node_t node, void* arg);
static int dumpMinorDeviceNode (di_node_t node, di_minor_t minor, void *arg);
static void walkDiskNodes (const char* driver, di_node_t node);
static void resetDDIInfo (void);
static void resetMnttabEntries (void);
static int guessPropertyType(uchar_t *data, int len);
static int readDir (const char* diskPath, const char* svr4Path, char* partitionName);
static void setProperties (di_node_t node, DiskSCSIInfo* scsiInfo);
static void setTrickySCSIInfo (DiskSCSIInfo* scsiInfo);
static void setKIODiskStat (DiskSCSIInfo* scsiInfo);
static void updateDDIInfo (di_node_t node, const char* driver);
static DiskSCSIInfo* findDiskSCSIInfo (const char* ddiDeviceName);
static DiskSCSIInfo* addNewDiskSCSIInfo (size_t currentSize);
static MnttabInfo* addNewMnttabInfo (size_t currentSize);
static DDIInfo* findDDIInfo (const char* ddiDeviceName);
static int findInstance (int instance);
static DiskSCSIInfo* findSCSISlice (const char* sliceName);
static MnttabInfo* findMnttabinfo (int32_t major, int32_t minor);
static void addInstance (int instance);
static void resetInstanceInfo (void);

extern int errno;

void
setUTSName (void)
{
  if (uname (&(utsName.uName)) < 0)
  {
    perror("uname"), strerror(errno);
    utsName.ok = 0;
  }
  else
    utsName.ok = 1;
}

void
setDriverName (void)
{
  char* ptr;
  int version;

  if (!utsName.ok)
    setUTSName();

  if (utsName.ok)
  {
    ptr = strchr (utsName.uName.release, '.');
    if (ptr)
    {
      ++ptr;
      version = atoi(ptr);
      if (version == 10)
        (void) strcpy (driver, "ssd");
      else
        (void) strcpy (driver, "sd");
      hasDriver = 1;
    }
  }
}

char*
getSCSIDevice (const char* cmd)
{
  char* ptr;
  char format[32];
  
  (void) memset (device, '\0', sizeof (device));
  (void) memset (format, '\0', 32);
  (void) sprintf (format, "scsi%%%us", (uint32_t) MAXPATHLEN);
  (void) sscanf (cmd, format, device);
  
  ptr = (char *) rindex (device, '/');
  if (ptr)
  {
    *ptr = '\0';
    return device;
  }
  return 0L;
}

int
getMnttabInfo (void)
{
    struct extmnttab mp;
    FILE* fp = 0L;
    uint32_t i;
    
    if ((fp = fopen (MNTTAB, "r")) == (FILE *) 0L)
    {
        perror("fopen"), strerror(errno);
        return 0L;
    }
    
    (void) memset ((void *) &mp, '\0', sizeof (struct extmnttab));
    resetMnttabEntries();
    
    while (0 == getextmntent (fp, &mp, MNT_LINE_MAX))
    {
        if ((0 == strcmp (mp.mnt_fstype, "ufs")) || (0 == strcmp (mp.mnt_special, "swap")))
        {
            if (0L != addNewMnttabInfo (numberOfMnttabEntries))
            {
              (void) strcpy (mnttabInfoPtr[numberOfMnttabEntries].device, mp.mnt_special);
              (void) strcpy (mnttabInfoPtr[numberOfMnttabEntries].mountPoint, mp.mnt_mountp);
              (void) strcpy (mnttabInfoPtr[numberOfMnttabEntries].fsType, mp.mnt_fstype);
              mnttabInfoPtr[numberOfMnttabEntries].major = mp.mnt_major;
              mnttabInfoPtr[numberOfMnttabEntries].minor = mp.mnt_minor;
              ++numberOfMnttabEntries;
            }
        }
    }
    resetmnttab (fp);
    fclose (fp);
    return (numberOfMnttabEntries ? 0 : -1);
}

int
updateSCSIStat (void)
{
    if (getSCSIStatInfo() > 0)
        return 0;
    
    return -1;
}

int
getSCSIStatInfo (void)
{
    if (getMnttabInfo() < 0)
        return -1;
    if (updateDeviceInfo(driver) != 0)
        return -1;
    if (updateKStatList() != 0)
        return -1;
    return 0;
}


void
initSCSIStat (struct SensorModul* sm)
{
  char monitor[MAXPATHLEN + (16 * sizeof(char))];
  uint32_t i;

#if 0
  if (0 == hasDriver)
  {
    setUTSName();
    setDriverName();

    if (0 == hasDriver)
    return;
  }
#endif

  DiskStatSM = sm;
  updateSCSIStat();
  
  registerMonitor("scsi/list", "listview", printSCSIStat, printSCSIStatInfo, DiskStatSM);
  
  for (i = 0; i < numSCSIInfo; ++i)
  {
      (void) snprintf (monitor, sizeof(monitor), "scsi%s/bread", scsiInfoArray[i].svr4Device);
      registerMonitor (monitor, "integer", printSCSIStatBytesRead, printSCSIStatBytesReadInfo, DiskStatSM);
      (void) snprintf (monitor, sizeof(monitor), "scsi%s/bwritten", scsiInfoArray[i].svr4Device);
      registerMonitor (monitor, "integer", printSCSIStatBytesWritten, printSCSIStatBytesWrittenInfo, DiskStatSM);
      (void) snprintf (monitor, sizeof(monitor), "scsi%s/nreads", scsiInfoArray[i].svr4Device);
      registerMonitor (monitor, "integer", printSCSIStatNumReadOps, printSCSIStatNumReadOpsInfo, DiskStatSM);
      (void) snprintf (monitor, sizeof(monitor), "scsi%s/nwrites", scsiInfoArray[i].svr4Device);
      registerMonitor (monitor, "integer", printSCSIStatNumWriteOps, printSCSIStatNumWriteOpsInfo, DiskStatSM);
  }
}


void
exitSCSIStat (void)
{
  char monitor[MAXPATHLEN + 12 * sizeof(char)];
  uint32_t i;

  removeMonitor("scsi/list");
  
  for (i = 0; i < numSCSIInfo; ++i)
  {
      (void) snprintf (monitor, sizeof(monitor), "scsi%s/bread", scsiInfoArray[i].svr4Device);
      removeMonitor (monitor);
      (void) snprintf (monitor, sizeof(monitor), "scsi%s/bwritten", scsiInfoArray[i].svr4Device);
      removeMonitor (monitor);
      (void) snprintf (monitor, sizeof(monitor), "scsi%s/nreads", scsiInfoArray[i].svr4Device);
      removeMonitor (monitor);
      (void) snprintf (monitor, sizeof(monitor), "scsi%s/nwrites", scsiInfoArray[i].svr4Device);
      removeMonitor (monitor);
  }
  free (scsiInfoArray);
  scsiInfoArray = 0L;
}

void
checkSCSIStat (void)
{
    struct extmnttab mp;
    unsigned int ctr = 0;
    int ret = 0;
    FILE* fp = 0L;
    uint32_t i;
    
    if ((fp = fopen (MNTTAB, "r")) == (FILE *) 0L)
    {
        perror("fopen"), strerror(errno);
        return;
    }
    (void) memset ((void *) &mp, '\0', sizeof (struct extmnttab));
    
    while ((ret = getextmntent (fp, &mp, MNT_LINE_MAX)) == 0)
    {
        if ((0 == strcmp (mp.mnt_fstype, "ufs")) || (0 == strcmp (mp.mnt_special, "swap")))
            ++ctr;
    }
    resetmnttab (fp);
    fclose (fp);
    
    if (ctr != numberOfMnttabEntries)
    {
        exitSCSIStat();
        initSCSIStat (DiskStatSM);
    }
}

void
fillDefaultValues (DiskSCSIInfo* dp)
{
  dp->instance = 0;
  dp->kid = 0;
  dp->lun = 0;
  dp->target = 0;
  dp->dev = 0x0;
  dp->major = 0;
  dp->minor = 0;
  (void) memset (dp->kstatName, '\0', sizeof (dp->kstatName));
  (void) memset (dp->kstatClass, '\0', sizeof (dp->kstatClass));
  (void) memset (dp->vendorName, '\0', sizeof (dp->vendorName));
  (void) memset (dp->productId, '\0', sizeof (dp->productId));
  (void) memset (dp->revision, '\0', sizeof (dp->revision));
  (void) memset (dp->ddiDeviceName, '\0', sizeof (dp->ddiDeviceName));
  (void) memset (dp->specType, '\0', sizeof (dp->specType));
  (void) memset (dp->nodeType, '\0', sizeof (dp->nodeType));
  (void) memset (dp->svr4Device, '\0', sizeof (dp->svr4Device));
  (void) memset (dp->scsiSlice, '\0', sizeof (dp->scsiSlice));
  (void) memset (dp->diMinorName, '\0', sizeof (dp->diMinorName));
  (void) memset (&(dp->kstatIO), '\0', sizeof (kstat_io_t));
}

void
copyDiskSCSIInfo (const kstat_t* ksp, const kstat_io_t* kip, DiskSCSIInfo* dp)
{
    if ((0L == ksp) || (0L == dp))
        return;
    
    dp->kid = ksp->ks_kid;
    (void) strcpy (dp->kstatName, ksp->ks_name);
    (void) strcpy (dp->kstatClass, ksp->ks_class);
    (void) memcpy (&(dp->kstatIO), kip, sizeof (kstat_io_t));
}

void
resetMemory (void)
{
    if (scsiInfoArray != 0L)
    {
        free (scsiInfoArray);
        scsiInfoArray = 0L;
        numSCSIInfo = 0;
    }
}

void
resetMnttabEntries (void)
{
    if (mnttabInfoPtr != 0L)
    {
        free (mnttabInfoPtr);
        mnttabInfoPtr = 0L;
        numberOfMnttabEntries = 0;
    }
}


int
updateDeviceAlias (void)
{
  char file[MAXPATHLEN];
  di_node_t rootNode;
  di_node_t node;
  di_minor_t minor = DI_MINOR_NIL;
  char tmpnode[MAXPATHLEN];
  char* phys_path;
  char* minor_name;
  int instance;

  resetDDIInfo();
  resetInstanceInfo();
  
  if ((rootNode = di_init("/", DINFOCPYALL)) == DI_NODE_NIL)
  {
      perror("di_init"), strerror(errno);
      return -1;
  }
  walkDiskNodes (driver, rootNode);
  node = di_drv_first_node(driver, rootNode);
  while (node != DI_NODE_NIL)
  {
      updateDDIInfo (node, driver);
      (void) di_walk_minor(node, NULL, 0, (void *) node, dumpMinorDeviceNode);
      node = di_drv_next_node(node);
  }
  di_fini(rootNode);
  return 0;
}

void
walkDiskNodes (const char* driver, di_node_t node)
{
    node = di_drv_first_node(driver, node);
    if (node == DI_NODE_NIL)
        return;

    updateDDIInfo (node, driver);
    node = di_drv_next_node(node);
    
    while (node != DI_NODE_NIL)
    {
        updateDDIInfo (node, driver);
        node = di_drv_next_node(node);
    }
}


int
updateKStatList (void)
{
    DiskSCSIInfo* diskInfoPtr = 0L;
    kstat_ctl_t* kc = 0L;
    kstat_t* ks = 0L;
    kstat_io_t kio;
    uint32_t i;
    
    if ((kc = kstat_open()) == (kstat_ctl_t *) NULL)
    {
        perror("kstat_open"), strerror(errno);
        return -1;
    }
    for (i = 0; i < numInstances; ++i)
    {
        if ((ks = kstat_lookup (kc, (char *) driver, instancePtr[i], NULL)) == (kstat_t *) NULL)
        {
            perror("kstat_lookup"), strerror(errno);
            continue;
        }
        
        while (kstat_chain_update(kc) != 0);
        
        do
        {
            if ((strcmp (ks->ks_module, driver) == 0) && (ks->ks_type == KSTAT_TYPE_IO))
            {
                if (kstat_read (kc, ks, &kio) == -1)
                {
                    perror("kstat_read"), strerror(errno);
                    kstat_close(kc);
                    return -1;
                }
                diskInfoPtr = findSCSISlice (ks->ks_name);
                if (0L == diskInfoPtr)
                {
                    while (kstat_chain_update(kc) != 0);
                    ks = ks->ks_next;
                    continue;
                }
                copyDiskSCSIInfo (ks, &kio, diskInfoPtr);
                while (kstat_chain_update(kc) != 0);
            }
            ks = ks->ks_next;
        } while (ks);
    }
    kstat_close(kc);
    return 0;
}

int
updateDeviceInfo (const char* driver)
{
    di_node_t rootNode;
    di_node_t node;
    di_prom_handle_t promHandle;
    int promProperties = 0;
    
    if ((rootNode = di_init("/", DINFOCPYALL)) == DI_NODE_NIL)
    {
        perror("di_init"), strerror(errno);
        return -1;
    }
    if ((0 == getuid()) || (0 == geteuid()))
    {
        ++promProperties;
        if (promProperties && (promHandle = di_prom_init()) == DI_PROM_HANDLE_NIL)
        {
            perror("di_prom_init"), strerror(errno);
            di_fini(rootNode);
            return -1;
        }
    }
    node = di_drv_first_node(driver, rootNode);
    while (node != DI_NODE_NIL)
    {
        (void) di_walk_minor(node, NULL, 0, (void *) node, dumpMinorDeviceNode);
        node = di_drv_next_node(node);
    }

    if (promProperties)
      di_prom_fini(promHandle);

    di_fini(rootNode);
    return 0;
}

void
resetInstanceInfo (void)
{
    if (instancePtr)
    {
        free (instancePtr);
        instancePtr = 0L;
    }
}


void
resetDDIInfo (void)
{
    if (ddiInfoPtr)
    {
        free (ddiInfoPtr);
        ddiInfoPtr = 0L;
    }
}

void
updateDDIInfo (di_node_t node, const char* driver)
{
  int instance;
  char* physPath;

  if (di_minor_next(node, DI_MINOR_NIL) == DI_MINOR_NIL)
    return;

  instance = di_instance(node);
  physPath = di_devfs_path(node);
  
  ddiInfoPtr = (DDIInfo *) realloc (ddiInfoPtr, (numDDIInfo * sizeof(DDIInfo) + sizeof(DDIInfo)));
  if (0L == ddiInfoPtr)
    return;
  
  ddiInfoPtr[numDDIInfo].instance = instance;
  (void) strcpy (ddiInfoPtr[numDDIInfo].physicalPath, physPath);
  (void) strcpy (ddiInfoPtr[numDDIInfo].driver, driver);
  ++numDDIInfo;
}


int
dumpDeviceNode (di_node_t node, void* arg)
{
    char* name;
    di_prom_handle_t ph = (di_prom_handle_t)arg;
    
    name = di_devfs_path (node);

    if (di_minor_next (node, DI_MINOR_NIL) == DI_MINOR_NIL)
    {
        di_devfs_path_free(name);
        return DI_WALK_TERMINATE;
    }
    
    (void) di_walk_minor(node, NULL, 0, (void *) node, dumpMinorDeviceNode);
    di_devfs_path_free(name);
    return DI_WALK_CONTINUE;
}

int
dumpMinorDeviceNode (di_node_t node, di_minor_t minor, void *arg)
{
    char* name;
    char* specType;
    char* nodeType;
    char* physPath;
    char* minorName;
    dev_t dev;
    DiskSCSIInfo* scsiInfo;
    int instance;
    char svr4[MAXPATHLEN];
    char tmpNode[MAXPATHLEN];
    char file[MAXPATHLEN];
    char partition[MAXPATHLEN];

    name = di_devfs_path(node);
    
    if ((caddr_t) node != (caddr_t)arg)
        return DI_WALK_TERMINATE;
    
    instance = di_instance(node);
    physPath = di_devfs_path(node);
    minorName = di_minor_name(minor);
    (void) memset (partition, '\0', MAXPATHLEN);
    (void) memset (svr4, '\0', MAXPATHLEN);
    
    (void) sprintf (tmpNode, "%s%d", driver, instance);
    (void) strcpy (file, "/devices");
    (void) strcat (file, physPath);
    (void) strcat (file, ":");
    (void) strcat (file, minorName);
    
    specType = (di_minor_spectype(minor) == S_IFCHR) ? "char" : "block";
    nodeType = di_minor_nodetype(minor);
    
    if (di_minor_spectype(minor) == S_IFBLK)
        (void) strcpy (svr4, "/dev/dsk");
    else if (di_minor_spectype(minor) == S_IFCHR)
        (void) strcpy (svr4, "/dev/rdsk");
    else
    {
        di_devfs_path_free(name);
        return DI_WALK_TERMINATE;
    }
    if (0 == readDir(file, svr4, partition))
    {
        scsiInfo = findDiskSCSIInfo(file);
        if (0L == scsiInfo)
        {
            scsiInfo = addNewDiskSCSIInfo (numSCSIInfo);
            if ((0L == scsiInfo) || (0L == scsiInfoArray))
            { 
                di_devfs_path_free(name);
                return DI_WALK_TERMINATE;
            }
            
            fillDefaultValues (&(scsiInfoArray[numSCSIInfo]));
            scsiInfo = &(scsiInfoArray[numSCSIInfo]);
            ++numSCSIInfo;
        }
        scsiInfo->instance = instance;
        addInstance(scsiInfo->instance);
        scsiInfo->dev = di_minor_devt(minor);
        if (scsiInfo->dev != DDI_DEV_T_NONE)
        {
            scsiInfo->major = major(scsiInfo->dev);
            scsiInfo->minor = minor(scsiInfo->dev);
        }
        (void) strcpy (scsiInfo->ddiDeviceName, file);
        (void) strcpy (scsiInfo->svr4Device, partition);
        (void) strcpy (scsiInfo->specType, specType);
        (void) strcpy (scsiInfo->nodeType, ((nodeType) ? nodeType : ""));
        (void) strcpy (scsiInfo->diMinorName, di_minor_name(minor));
        (void) sprintf (scsiInfo->scsiSlice, "%s%d,%s", driver, instance, scsiInfo->diMinorName);
        setProperties (node, scsiInfo);
        setTrickySCSIInfo (scsiInfo);
    } 
    di_devfs_path_free(name);
    return DI_WALK_CONTINUE;
}

void
setProperties (di_node_t node, DiskSCSIInfo* scsiInfo)
{
    int rval, i, type;
    di_prop_t prop, next;
    char buf[256];
    int* intp;
    char* strp;
    char* ptr;
    uchar_t* bytep;
    
    if ((next = di_prop_next(node, DI_PROP_NIL)) == DI_PROP_NIL)
        return;

    while (next != DI_PROP_NIL)
    {
        prop = next;
        next = di_prop_next(node, prop);
        type = di_prop_type(prop);
        if (type == DI_PROP_TYPE_UNKNOWN)
        {
            rval = di_prop_bytes(prop, &bytep);
            type = guessPropertyType(bytep, rval);
        }
        if (type == DI_PROP_TYPE_INT)
        {
            if ((strcmp (di_prop_name(prop), "lun") == 0) || (strcmp (di_prop_name(prop), "target") == 0))
            {
                rval = di_prop_ints(prop, &intp);
                if (rval == -1)
                {
                    perror("di_prop_ints"), strerror(errno);
                    break;
                }
                (void) memset (buf, '\0', sizeof(buf));
                for (i = 0; i < rval; ++i)
                {
                    ptr = &buf[i];
                    (void) sprintf (ptr, "%d", intp[i]);
                }
                if (strcmp(di_prop_name(prop), "lun") == 0)
                    scsiInfo->lun = atoi(buf);
                else
                    scsiInfo->target = atoi(buf); 
            }
        }
        else if (type == DI_PROP_TYPE_STRING)
        {
            if ((strcmp (di_prop_name(prop), "inquiry-vendor-id") == 0) ||
                (strcmp (di_prop_name(prop), "inquiry-product-id") == 0) ||
                (strcmp (di_prop_name(prop), "inquiry-revision-id") == 0))
            {
                rval = di_prop_strings(prop, &strp);
                if (rval == -1)
                {
                    perror("di_prop_strings"), strerror(errno);
                    break;
                }
                if (strcmp (di_prop_name(prop), "inquiry-vendor-id") == 0)
                    (void) sprintf (scsiInfo->vendorName, "%s", strp);
                else if (strcmp (di_prop_name(prop), "inquiry-product-id") == 0)
                    (void) sprintf (scsiInfo->productId, "%s", strp);
                else
                    (void) sprintf (scsiInfo->revision, "%s", strp);
                strp += strlen(strp) + 1;
            }
        }
    }
}

void
setTrickySCSIInfo (DiskSCSIInfo* scsiInfo)
{
    if (0L == scsiInfo)
        return;
    
    if (0 == strcmp (scsiInfo->specType, "block"))
    {
        if (0 == strcmp (scsiInfo->diMinorName, "h"))
        {
            (void) sprintf (scsiInfo->kstatName, "%s%d", driver, scsiInfo->target);
            if (0 == strlen(scsiInfo->kstatClass))
                (void) strcpy (scsiInfo->kstatClass, "disk");
        }
        else
        {
            if (0 == strlen(scsiInfo->kstatClass))
                (void) strcpy (scsiInfo->kstatClass, "disk");
        }
    }
}

void
setKIODiskStat (DiskSCSIInfo* scsiInfo)
{
  uint32_t i;
  int target;

  if (0L == scsiInfo)
    return;

  target = scsiInfo->target;

  scsiInfo->kstatIO.reads = 0;
  scsiInfo->kstatIO.writes = 0;
  scsiInfo->kstatIO.nread = 0ULL;
  scsiInfo->kstatIO.nwritten = 0ULL;
}


int
guessPropertyType(uchar_t *data, int len)
{
    int i, c, slen, guess;
    
    if (len < 0)
        return (-1);
    else if (len == 0)
        return (DI_PROP_TYPE_BOOLEAN);
    
    slen = 0;
    guess = DI_PROP_TYPE_STRING;
    
    for (i = 0; i < len; i++)
    {
        c = (int)data[i];
        switch (c)
        {
            case 0: /* NUL */
                if (slen == 0)
                    guess = DI_PROP_TYPE_BYTE;
                else
                    slen = 0;
                break;
            default:
                if (!isprint(c))
                    guess = DI_PROP_TYPE_BYTE;
                else
                    slen++;
        }
        if (guess != DI_PROP_TYPE_STRING)
            break;
    }
    if ((guess == DI_PROP_TYPE_BYTE) && (len % sizeof(int) == 0))
        guess = DI_PROP_TYPE_INT;
    
    return (guess);
}

int
readDir (const char* diskPath, const char* svr4Path, char* partitionName)
{
    DIR* dirp;
    struct dirent *dp;
    struct stat stbuf;
    char fileName[MAXPATHLEN];
    char tempName[MAXPATHLEN];
    char dirDname[MAXPATHLEN];
    
    if ((dirp = opendir(svr4Path)) != NULL)
    {
        while ((dp = readdir(dirp)) != NULL)
        {
            (void) sprintf(tempName, "../..%s", diskPath);
            (void) sprintf(dirDname, "%s/%s", svr4Path, dp->d_name);
            stat (dirDname, &stbuf);

            if (0 == strcmp (svr4Path, "/dev/dsk"))
            {
                if (S_ISBLK(stbuf.st_mode))
                {
                    readlink(dirDname, fileName, sizeof(fileName));
                    if (strcmp(fileName, tempName) == 0)
                    {
                        (void) strcpy (partitionName, dirDname);
                        closedir(dirp);
                        return 0;
                    }
                }
            }
            else if (0 == strcmp (svr4Path, "/dev/rdsk"))
            {
                if (S_ISCHR(stbuf.st_mode))
                {
                    readlink(dirDname, fileName, sizeof(fileName));
                    if (strcmp(fileName, tempName) == 0)
                    {
                        (void) strcpy (partitionName, dirDname);
                        closedir(dirp);
                        return 0;
                    }
                }
            }
            (void) memset (fileName,'\0', MAXPATHLEN);
            (void) memset (tempName,'\0', MAXPATHLEN);
            (void) memset (dirDname,'\0', MAXPATHLEN);
        }
        closedir(dirp);
    }
    return(-1);
}

DiskSCSIInfo*
addNewDiskSCSIInfo (size_t currentSize)
{
  DiskSCSIInfo* ptr;
  DiskSCSIInfo* tmp;
  uint32_t i;

  ptr = (DiskSCSIInfo *) malloc ((1 + currentSize) * sizeof (DiskSCSIInfo));
  if (0 == ptr)
  {
    perror("malloc"), strerror(errno);
    return 0L;
  }

  if (0 == currentSize)
  {
    scsiInfoArray = ptr;
    return (&(scsiInfoArray[currentSize]));
  }

  for (i = 0; i < (uint32_t) currentSize; ++i)
    (void) memcpy (&(ptr[i]), &(scsiInfoArray[i]), sizeof (DiskSCSIInfo));

  tmp = scsiInfoArray;
  scsiInfoArray = ptr;
  free (tmp);
  return (&(scsiInfoArray[currentSize]));
}

MnttabInfo*
addNewMnttabInfo (size_t currentSize)
{
  MnttabInfo* ptr;
  MnttabInfo* tmp;
  uint32_t i;

  ptr = (MnttabInfo *) malloc ((1 + currentSize) * sizeof (MnttabInfo));
  if (0 == ptr)
  {
    perror("malloc"), strerror(errno);
    return 0L;
  }

  if (0 == currentSize)
  {
    mnttabInfoPtr = ptr;
    return (&(mnttabInfoPtr[currentSize]));
  }

  for (i = 0; i < (uint32_t) currentSize; ++i)
    (void) memcpy (&(ptr[i]), &(mnttabInfoPtr[i]), sizeof (MnttabInfo));

  tmp = mnttabInfoPtr;
  mnttabInfoPtr = ptr;
  free (tmp);
  return (&(mnttabInfoPtr[currentSize]));
}


DiskSCSIInfo*
findDiskSCSIInfo (const char* ddiDeviceName)
{
    uint32_t i;
    
    if (0L == ddiDeviceName)
        return 0L;
    
    for (i = 0; i < numSCSIInfo; ++i)
    {
        if (0 == strcmp (scsiInfoArray[i].ddiDeviceName, ddiDeviceName))
            return (&(scsiInfoArray[i]));
    }
    return 0L;
}

DDIInfo*
findDDIInfo (const char* ddiDeviceName)
{
    uint32_t i;

    if (0L == ddiDeviceName)
        return 0L;
    
    for (i = 0; i < numDDIInfo; ++i)
    {
        if (0 == strcmp (ddiInfoPtr[i].physicalPath, ddiDeviceName))
            return (&(ddiInfoPtr[i]));
    }
    return 0L;
}

int
findInstance (int instance)
{
    uint32_t i;
    
    for (i = 0; i < numInstances; ++i)
        if (instancePtr[i] == instance)
            return instancePtr[i];

    return -1;
}

DiskSCSIInfo*
findSCSISlice (const char* sliceName)
{
    uint32_t i;
    
    if (0L == sliceName)
        return 0L;
    
    for (i = 0; i < numSCSIInfo; ++i)
        if (strcmp (scsiInfoArray[i].scsiSlice, sliceName) == 0)
            return (&(scsiInfoArray[i]));

    return 0L;
}

MnttabInfo*
findMnttabInfo (int32_t major, int32_t minor)
{
    uint32_t i;
    
    for (i = 0; i < numberOfMnttabEntries; ++i)
    {
        if ((mnttabInfoPtr[i].major == major) &&
            (mnttabInfoPtr[i].minor == minor))
            return (&mnttabInfoPtr[i]);
    }
    return 0L;
}


void
addInstance (int instance)
{
    if (findInstance (instance) < 0)
    {
        instancePtr = (int *) realloc (instancePtr, sizeof(int) * numInstances + sizeof(int));
        if (instancePtr)
        {
            instancePtr[numInstances] = instance;
            ++numInstances;
        }
    }
}

/*
 * public interface
 */

void
printSCSIStat (const char* cmd)
{
    uint32_t i;
    const char* ptr;
    char kstatName[16];
    char kstatClass[64];
    
    for (i = 0; i < numSCSIInfo; i++)
    {
      if ((0 == scsiInfoArray[i].kid) && !(0 == strcmp (scsiInfoArray[i].diMinorName, "h")))
          continue;

      if (0 == strcmp (scsiInfoArray[i].specType, "block"))
      {
        if (0 == strcmp (scsiInfoArray[i].diMinorName, "h"))
          setKIODiskStat (&scsiInfoArray[i]);

        (void) fprintf (CurrentClient, "0x%x\t%ld\t%ld\t%i\t%i\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%u\t%llu\t%u\t%llu\n",
                        scsiInfoArray[i].dev,
                        scsiInfoArray[i].major,
                        scsiInfoArray[i].minor,
                        scsiInfoArray[i].lun,
                        scsiInfoArray[i].target,
                        scsiInfoArray[i].kstatName,
                        scsiInfoArray[i].kstatClass,
                        scsiInfoArray[i].vendorName,
                        scsiInfoArray[i].productId,
                        scsiInfoArray[i].revision,
                        scsiInfoArray[i].ddiDeviceName,
                        scsiInfoArray[i].specType,
                        scsiInfoArray[i].nodeType,
                        scsiInfoArray[i].svr4Device,
                        scsiInfoArray[i].scsiSlice,
                        scsiInfoArray[i].kstatIO.reads,
                        scsiInfoArray[i].kstatIO.nread,
                        scsiInfoArray[i].kstatIO.writes,
                        scsiInfoArray[i].kstatIO.nwritten);
      }
    }
    (void) fprintf (CurrentClient, "\n");
}

void
printSCSIStatInfo (const char* cmd)
{
    (void) fprintf(CurrentClient, "Device\tMajor\tMinor\tLUn\tTarget\tName\tClass\tVendor\tProductID\tRevision\tDDIDeviceName\tSpec\tNode\tSVR4-Style\tSlice\tReadOperations\tBytesRead\tWriteOperations\tBytesWritten\ns\tD\tD\tD\tD\ts\ts\ts\ts\ts\ts\ts\ts\ts\ts\tD\tD\tD\tD\n");
}

void
printSCSIStatNumReadOps (const char* cmd)
{
  uint32_t i;

  char* scsiDevice = getSCSIDevice (cmd);
  if (scsiDevice)
  {
      for (i = 0; i < numSCSIInfo; i++)
      {
          if (0 == strcmp (scsiDevice, scsiInfoArray[i].svr4Device))
              (void) fprintf (CurrentClient, "%u\n", scsiInfoArray[i].kstatIO.reads);
      }
  }
  (void) fprintf (CurrentClient, "\n");
}

void
printSCSIStatNumReadOpsInfo (const char* cmd)
{
    (void) fprintf (CurrentClient, "Read Operations\t0\t-\tReads\n");
}

void
printSCSIStatNumWriteOps (const char* cmd)
{
  uint32_t i;
  
  char* scsiDevice = getSCSIDevice (cmd);
  if (scsiDevice)
  {
      for (i = 0; i < numSCSIInfo; i++)
      {
          if (0 == strcmp (scsiDevice, scsiInfoArray[i].svr4Device))
              (void) fprintf (CurrentClient, "%u\n", scsiInfoArray[i].kstatIO.writes);
      }
  }
  (void) fprintf (CurrentClient, "\n");
}

void
printSCSIStatNumWriteOpsInfo (const char* cmd)
{
  (void) fprintf (CurrentClient, "Write Operations\t0\t-\tWrites\n");
}

void
printSCSIStatBytesRead (const char* cmd)
{
    uint32_t i;
    
    char* scsiDevice = getSCSIDevice (cmd);
    if (scsiDevice)
    {
        for (i = 0; i < numSCSIInfo; i++)
        {
            if (0 == strcmp (scsiDevice, scsiInfoArray[i].svr4Device))
                (void) fprintf (CurrentClient, "%llu\n", scsiInfoArray[i].kstatIO.nread);
        }
    }
}

void
printSCSIStatBytesReadInfo (const char* cmd)
{
    (void) fprintf (CurrentClient, "Bytes Read\t0\t-\tBytes\n");
}

void
printSCSIStatBytesWritten (const char* cmd)
{
    uint32_t i;
    
    char* scsiDevice = getSCSIDevice (cmd);
    if (scsiDevice)
    {
        for (i = 0; i < numSCSIInfo; i++)
        {
            if (0 == strcmp (scsiDevice, scsiInfoArray[i].svr4Device))
                (void) fprintf (CurrentClient, "%llu\n", scsiInfoArray[i].kstatIO.nwritten);
        }
    }
}

void
printSCSIStatBytesWrittenInfo (const char* cmd)
{
    (void) fprintf (CurrentClient, "Bytes Written\t0\t-\tBytes\n");
}

// vim: ts=2 sw=2 et


