/*
  KSysGuard, the KDE System Guard
  
  Copyright (c) 2005 Stefan Teleman <stefan.teleman@gmail.com>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of version 2 of the GNU General Public
  License as published by the Free Software Foundation.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
*/

#include <config.h>

#include "Command.h"
#include "ccont.h"
#include "SCSIStatCommon.h"
#include "SCSIStatSSD.h"
#include "ksysguardd.h"

static const char* driverName = "ssd";
static char ssdDevice[MAXPATHLEN + sizeof(char)];
static struct SensorModul* DiskStatSM = 0L;
static char device[MAXPATHLEN + sizeof(char)];
static DDIInfoCtrl* ddiInfoCtrlPtr = 0L;
static DiskSCSIInfoCtrl* diskSCSIInfoCtrlPtr = 0L;
static MnttabInfoCtrl* mnttabInfoCtrlPtr = 0L;
static InstanceCtrl* instanceCtrlPtr = 0L;
static uint32_t ssdInit = 0;
static struct timeval tv;

extern int errno;

static void initMemory (void);
static void resetMemory (void);
static char* getSSDSCSIDevice (const char* cmd);
static int getSSDSCSIStatInfo (void);

char*
getSSDSCSIDevice (const char* cmd)
{
    char* ptr;
    char format[32];
    
  (void) memset (ssdDevice, '\0', sizeof (ssdDevice));
  (void) memset (format, '\0', 32);
  (void) sprintf (format, "scsissd%%%us", (uint32_t) MAXPATHLEN);
  (void) sscanf (cmd, format, ssdDevice);
  
  ptr = (char *) rindex (ssdDevice, '/');
  if (ptr)
  {
      *ptr = '\0';
      return ssdDevice;
  }
  return 0L;
}

int
updateSSDSCSIStat (void)
{
   if (0 == getSSDSCSIStatInfo())
     return 0;

    return -1;
}

int
getSSDSCSIStatInfo (void)
{
  if (getMnttabInfo (mnttabInfoCtrlPtr) < 0)
      return -1;
  if (updateDeviceInfo (&diskSCSIInfoCtrlPtr, &instanceCtrlPtr, driverName) != 0)
      return -1;
  if (updateKStatList (diskSCSIInfoCtrlPtr, instanceCtrlPtr, driverName) != 0)
      return -1;
  return 0;
}


void
initSSDSCSIStat (struct SensorModul* sm)
{
  char monitor[MAXPATHLEN + 16 * sizeof(char)];
  uint32_t i;

  DiskStatSM = sm;
  initMemory();
  updateSSDSCSIStat();

  if (0L == diskSCSIInfoCtrlPtr)
    return;

  if (0 == diskSCSIInfoCtrlPtr->numSCSIInfo)
    return;

  registerMonitor("scsissd/list", "listview", printSSDSCSIStat, printSSDSCSIStatInfo, DiskStatSM);
  
  for (i = 0; i < diskSCSIInfoCtrlPtr->numSCSIInfo; ++i)
  {
      (void) snprintf (monitor, sizeof(monitor), "scsissd%s/bread", diskSCSIInfoCtrlPtr->scsiInfoPtr[i].svr4Device);
      registerMonitor (monitor, "integer", printSSDSCSIStatBytesRead, printSSDSCSIStatBytesReadInfo, DiskStatSM);
      (void) snprintf (monitor, sizeof(monitor), "scsissd%s/bwritten", diskSCSIInfoCtrlPtr->scsiInfoPtr[i].svr4Device);
      registerMonitor (monitor, "integer", printSSDSCSIStatBytesWritten, printSSDSCSIStatBytesWrittenInfo, DiskStatSM);
      (void) snprintf (monitor, sizeof(monitor), "scsissd%s/nreads", diskSCSIInfoCtrlPtr->scsiInfoPtr[i].svr4Device);
      registerMonitor (monitor, "integer", printSSDSCSIStatNumReadOps, printSSDSCSIStatNumReadOpsInfo, DiskStatSM);
      (void) snprintf (monitor, sizeof(monitor), "scsissd%s/nwrites", diskSCSIInfoCtrlPtr->scsiInfoPtr[i].svr4Device);
      registerMonitor (monitor, "integer", printSSDSCSIStatNumWriteOps, printSSDSCSIStatNumWriteOpsInfo, DiskStatSM);
  }
}


void
exitSSDSCSIStat (void)
{
  char monitor[MAXPATHLEN + 16 * sizeof(char)];
  uint32_t i;

  if (0L == diskSCSIInfoCtrlPtr)
    return;

  if (0 == diskSCSIInfoCtrlPtr->numSCSIInfo)
    return;

  removeMonitor("scsisd/list");

  for (i = 0; i < diskSCSIInfoCtrlPtr->numSCSIInfo; ++i)
  {
      (void) snprintf (monitor, sizeof(monitor), "scsissd%s/bread", diskSCSIInfoCtrlPtr->scsiInfoPtr[i].svr4Device);
      removeMonitor (monitor);
      (void) snprintf (monitor, sizeof(monitor), "scsissd%s/bwritten", diskSCSIInfoCtrlPtr->scsiInfoPtr[i].svr4Device);
      removeMonitor (monitor);
      (void) snprintf (monitor, sizeof(monitor), "scsissd%s/nreads", diskSCSIInfoCtrlPtr->scsiInfoPtr[i].svr4Device);
      removeMonitor (monitor);
      (void) snprintf (monitor, sizeof(monitor), "scsissd%s/nwrites", diskSCSIInfoCtrlPtr->scsiInfoPtr[i].svr4Device);
      removeMonitor (monitor);
  }
  resetMemory();
}


void
checkSSDSCSIStat (void)
{
    struct extmnttab mp;
    unsigned int ctr = 0;
    int ret = 0;
    FILE* fp = 0L;
    uint32_t i;
    
    if ((fp = fopen (MNTTAB, "r")) == (FILE *) 0L)
    {
        perror("fopen"), strerror(errno);
        return;
    }
    (void) memset ((void *) &mp, '\0', sizeof (struct extmnttab));
    
    while ((ret = getextmntent (fp, &mp, MNT_LINE_MAX)) == 0)
    {
        if ((0 == strcmp (mp.mnt_fstype, "ufs")) || (0 == strcmp (mp.mnt_special, "swap")))
            ++ctr;
    }
    resetmnttab (fp);
    fclose (fp);
    
    if (ctr != mnttabInfoCtrlPtr->numberOfMnttabEntries)
    {
        exitSDSCSIStat();
        initSDSCSIStat (DiskStatSM);
    }
}

void
initMemory (void)
{
    if (0 == ssdInit)
    {
        if (0L == ddiInfoCtrlPtr)
        {
            ddiInfoCtrlPtr = (DDIInfoCtrl *) malloc (sizeof (DDIInfoCtrl));
            ddiInfoCtrlPtr->ddiInfoPtr = 0L;
            uint32_t numDDIInfo = 0;
        }
        
        if (0L == diskSCSIInfoCtrlPtr)
        {
            diskSCSIInfoCtrlPtr = (DiskSCSIInfoCtrl *) malloc (sizeof (DiskSCSIInfoCtrl));
            diskSCSIInfoCtrlPtr->scsiInfoPtr = 0L;
            diskSCSIInfoCtrlPtr->numSCSIInfo = 0;
        }
        
        if (0L == mnttabInfoCtrlPtr)
        {
            mnttabInfoCtrlPtr = (MnttabInfoCtrl *) malloc (sizeof (MnttabInfoCtrl));
            mnttabInfoCtrlPtr->mnttabInfoPtr = 0L;
            mnttabInfoCtrlPtr->numberOfMnttabEntries = 0;
        }
        
        if (0L == instanceCtrlPtr)
        {
            instanceCtrlPtr = (InstanceCtrl *) malloc (sizeof (InstanceCtrl));
            instanceCtrlPtr->instancePtr = 0L;
            instanceCtrlPtr->numInstances = 0;
        }
        ssdInit = 1;
    }
}

void
resetMemory (void)
{
  if (ddiInfoCtrlPtr)
  {
    if (ddiInfoCtrlPtr->ddiInfoPtr)
    {
      free (ddiInfoCtrlPtr->ddiInfoPtr);
      ddiInfoCtrlPtr->ddiInfoPtr = 0L;
    }
    free (ddiInfoCtrlPtr);
    ddiInfoCtrlPtr = 0L;
  }

  if (diskSCSIInfoCtrlPtr)
  {
    if (diskSCSIInfoCtrlPtr->scsiInfoPtr)
    {
      free (diskSCSIInfoCtrlPtr->scsiInfoPtr);
      diskSCSIInfoCtrlPtr->scsiInfoPtr = 0L;
    }
    free (diskSCSIInfoCtrlPtr);
    diskSCSIInfoCtrlPtr = 0L;
  }

  if (mnttabInfoCtrlPtr)
  {
    if (mnttabInfoCtrlPtr->mnttabInfoPtr)
    {
      free (mnttabInfoCtrlPtr->mnttabInfoPtr);
      mnttabInfoCtrlPtr->mnttabInfoPtr = 0L;
    }
    free (mnttabInfoCtrlPtr);
    mnttabInfoCtrlPtr = 0L;
  }

  if (instanceCtrlPtr)
  {
    if (instanceCtrlPtr->instancePtr)
    {
      free (instanceCtrlPtr->instancePtr);
      instanceCtrlPtr->instancePtr = 0L;
    }
    free (instanceCtrlPtr);
    instanceCtrlPtr = 0L;
  }
  ssdInit = 0;
}


/*
 * public interface
 */

void
printSSDSCSIStat (const char* cmd)
{
    uint32_t i;
    const char* ptr;
    char kstatName[16];
    char kstatClass[64];

    if (0L == diskSCSIInfoCtrlPtr)
        return;

    if (0L == diskSCSIInfoCtrlPtr->scsiInfoPtr)
        return;
    
    for (i = 0; i < diskSCSIInfoCtrlPtr->numSCSIInfo; i++)
    {
        if ((0 == diskSCSIInfoCtrlPtr->scsiInfoPtr[i].kid) && !(0 == strcmp (diskSCSIInfoCtrlPtr->scsiInfoPtr[i].diMinorName, "h")))
            continue;
        
        if (0 == strcmp (diskSCSIInfoCtrlPtr->scsiInfoPtr[i].specType, "block"))
      {
        if (0 == strcmp (diskSCSIInfoCtrlPtr->scsiInfoPtr[i].diMinorName, "h"))
          setKIODiskStat (&(diskSCSIInfoCtrlPtr->scsiInfoPtr[i]));

        (void) fprintf (CurrentClient, "0x%x\t%ld\t%ld\t%i\t%i\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%u\t%llu\t%u\t%llu\n",
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].dev,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].major,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].minor,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].lun,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].target,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].kstatName,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].kstatClass,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].vendorName,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].productId,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].revision,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].ddiDeviceName,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].specType,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].nodeType,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].svr4Device,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].scsiSlice,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].kstatIO.reads,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].kstatIO.nread,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].kstatIO.writes,
                        diskSCSIInfoCtrlPtr->scsiInfoPtr[i].kstatIO.nwritten);
      }
    }
    (void) fprintf (CurrentClient, "\n");
}

void
printSSDSCSIStatInfo (const char* cmd)
{
    (void) fprintf(CurrentClient, "Device\tMajor\tMinor\tLUn\tTarget\tName\tClass\tVendor\tProductID\tRevision\tDDIDeviceName\tSpec\tNode\tSVR4-Style\tSlice\tReadOperations\tBytesRead\tWriteOperations\tBytesWritten\ns\tD\tD\tD\tD\ts\ts\ts\ts\ts\ts\ts\ts\ts\ts\tD\tD\tD\tD\n");
}


void
printSSDSCSIStatNumReadOps (const char* cmd)
{
  uint32_t i;

  char* scsiDevice = getSSDSCSIDevice (cmd);
  if (scsiDevice)
  {
      for (i = 0; i < diskSCSIInfoCtrlPtr->numSCSIInfo; i++)
      {
          if (0 == strcmp (scsiDevice, diskSCSIInfoCtrlPtr->scsiInfoPtr[i].svr4Device))
              (void) fprintf (CurrentClient, "%u\n", diskSCSIInfoCtrlPtr->scsiInfoPtr[i].kstatIO.reads);
      }
  }
  (void) fprintf (CurrentClient, "\n");
}

void
printSSDSCSIStatNumReadOpsInfo (const char* cmd)
{
    (void) fprintf (CurrentClient, "Read Operations\t0\t-\tReads\n");
}

void
printSSDSCSIStatNumWriteOps (const char* cmd)
{
  uint32_t i;
  
  char* scsiDevice = getSSDSCSIDevice (cmd);
  if (scsiDevice)
  {
      for (i = 0; i < diskSCSIInfoCtrlPtr->numSCSIInfo; i++)
      {
          if (0 == strcmp (scsiDevice, diskSCSIInfoCtrlPtr->scsiInfoPtr[i].svr4Device))
              (void) fprintf (CurrentClient, "%u\n", diskSCSIInfoCtrlPtr->scsiInfoPtr[i].kstatIO.writes);
      }
  }
  (void) fprintf (CurrentClient, "\n");
}

void
printSSDSCSIStatNumWriteOpsInfo (const char* cmd)
{
  (void) fprintf (CurrentClient, "Write Operations\t0\t-\tWrites\n");
}

void
printSSDSCSIStatBytesRead (const char* cmd)
{
    uint32_t i;
    
    char* scsiDevice = getSSDSCSIDevice (cmd);
    if (scsiDevice)
    {
        for (i = 0; i < diskSCSIInfoCtrlPtr->numSCSIInfo; i++)
        {
            if (0 == strcmp (scsiDevice, diskSCSIInfoCtrlPtr->scsiInfoPtr[i].svr4Device))
                (void) fprintf (CurrentClient, "%llu\n", diskSCSIInfoCtrlPtr->scsiInfoPtr[i].kstatIO.nread);
        }
    }
}

void
printSSDSCSIStatBytesReadInfo (const char* cmd)
{
    (void) fprintf (CurrentClient, "Bytes Read\t0\t-\tBytes\n");
}

void
printSSDSCSIStatBytesWritten (const char* cmd)
{
    uint32_t i;
    
    char* scsiDevice = getSSDSCSIDevice (cmd);
    if (scsiDevice)
    {
        for (i = 0; i < diskSCSIInfoCtrlPtr->numSCSIInfo; i++)
        {
            if (0 == strcmp (scsiDevice, diskSCSIInfoCtrlPtr->scsiInfoPtr[i].svr4Device))
                (void) fprintf (CurrentClient, "%llu\n", diskSCSIInfoCtrlPtr->scsiInfoPtr[i].kstatIO.nwritten);
        }
    }
}

void
printSSDSCSIStatBytesWrittenInfo (const char* cmd)
{
    (void) fprintf (CurrentClient, "Bytes Written\t0\t-\tBytes\n");
}

// vim: ts=2 sw=2 et


